"""Scenario"""

from __future__ import annotations

from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import List, Optional, Union

from .case import Case, CaseResult
from .context import Context
from .description import Description
from .status import (
    Status, StatusedMixin, StatusedSequence, collect_statused, merge_statuses
)
from .verification import Verification, collect


@dataclass(frozen=True)
class ScenarioResult(StatusedMixin):
    label: Optional[str] = None
    message: Optional[str] = None
    conditions: Verification = field(default_factory=Verification)
    cases: StatusedSequence[CaseResult] = field(
        default_factory=StatusedSequence,
    )
    subscenarios: StatusedSequence[ScenarioResult] = field(
        default_factory=StatusedSequence,
    )


class RunningScenarioTask:

    def __init__(
        self, label: Optional[str],
        conditions: Verification,
        cases: Future,
        subscenarios: List[ScenarioTask],
    ):
        self._label = label
        self._conditions = conditions
        self._cases = cases
        self._subscenarios = subscenarios

    def result(self) -> ScenarioResult:
        cases = self._cases.result()
        subscenarios = collect_statused(s.result() for s in self._subscenarios)
        status = merge_statuses(cases.status, subscenarios.status)
        return ScenarioResult(
            label=self._label,
            status=status,
            conditions=self._conditions,
            cases=cases,
            subscenarios=subscenarios,
        )


class StaticScenarioTask:

    def __init__(self, result: ScenarioResult):
        self._result = result

    def result(self) -> ScenarioResult:
        return self._result


ScenarioTask = Union[RunningScenarioTask, StaticScenarioTask]


class Scenario:

    def __init__(
        self,
        label: Optional[str] = None,
        conditions: List[Description] = [],
        cases: List[Case] = [],
        subscenarios: List[Scenario] = [],
    ):
        self._label = label
        self._conditions = conditions
        self._cases = cases
        self._subscenarios = subscenarios

    def run(
        self,
        base_url: str,
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
    ) -> ScenarioResult:
        with ThreadPoolExecutor(1) as executor:
            return self.submit(
                executor,
                base_url=base_url,
                retry=retry,
                delay=delay,
                timeout=timeout,
            ).result()

    def submit(
        self,
        executor: ThreadPoolExecutor,
        base_url: str,
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
    ) -> ScenarioTask:
        context = Context(base_url=base_url)
        context_analyzer = context.analyze()
        conditions = collect(
            condition(context_analyzer) for condition in self._conditions
        )
        if conditions.status == Status.FAILURE:
            return StaticScenarioTask(ScenarioResult(
                label=self._label,
                status=Status.FAILURE,
                conditions=conditions,
            ))
        if conditions.status == Status.UNSTABLE:
            return StaticScenarioTask(ScenarioResult(
                label=self._label,
                status=Status.SKIPPED,
                conditions=conditions,
            ))

        cases = executor.submit(
            self._run_cases,
            base_url=base_url,
            retry=retry,
            delay=delay,
            timeout=timeout,
        )
        subscenarios = [
            subscenario.submit(
                executor,
                base_url=base_url,
                retry=retry,
                delay=delay,
                timeout=timeout,
            )
            for subscenario in self._subscenarios
        ]
        return RunningScenarioTask(
            label=self._label,
            conditions=conditions,
            cases=cases,
            subscenarios=subscenarios,
        )

    def _run_cases(
        self,
        base_url: str,
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
    ) -> StatusedSequence[CaseResult]:
        return collect_statused(
            case(base_url, timeout=timeout, retry=retry, delay=delay)
            for case in self._cases
        )



"""Response descriptions."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, List, Mapping, Optional

from .analysis import Analyzer, JsonAnalyzer
from .body_description import BodyDescription
from .description import Description, Predicate
from .status import Status, merge_statuses
from .verification import Verification


@dataclass(frozen=True)
class ResponseVerification:
    status: Status
    status_code: Verification
    headers: Verification
    body: Verification


class ResponseDescription:

    def __init__(
        self,
        status_code_predicates: List[Predicate] = [],
        headers_descriptions: List[Description] = [],
        body_description: Optional[BodyDescription] = None,
        analyze_headers:
            Callable[[Mapping[str, str]], Analyzer] = JsonAnalyzer,
    ):
        self._status_code_predicates = status_code_predicates
        self._headers_descriptions = headers_descriptions
        self._body_description = body_description
        self._analyze_headers = analyze_headers

    def verify(
        self,
        status_code: int,
        headers: Mapping[str, str],
        body: str,
        **kwargs: Any,
    ) -> ResponseVerification:
        """`**kwargs` will be delegated to descriptions."""
        status_code_verification = self._verify_status_code(
            status_code,
            **kwargs,
        )

        try:
            headers_verification = self._verify_headers(headers, **kwargs)
        except Exception as error:
            headers_verification = Verification.of_error(error)

        body_verification = Verification.skipped()
        if self._body_description:
            body_verification = self._body_description.verify(body, **kwargs)

        status = merge_statuses(
            status_code_verification.status,
            headers_verification.status,
            body_verification.status,
        )
        return ResponseVerification(
            status=status,
            status_code=status_code_verification,
            headers=headers_verification,
            body=body_verification,
        )

    @property
    def status_code_predicates(self) -> List[Predicate]:
        return self._status_code_predicates

    @property
    def headers_descriptions(self) -> List[Description]:
        return self._headers_descriptions

    @property
    def body_description(self) -> Optional[BodyDescription]:
        return self._body_description

    def _verify_status_code(
        self,
        code: int,
        **kwargs: Any,
    ) -> Verification:
        children = [
            predicate(code, **kwargs)
            for predicate in self._status_code_predicates
        ]
        status = merge_statuses(v.status for v in children)
        return Verification(status=status, children=children)

    def _verify_headers(
        self,
        header: Mapping[str, str],
        **kwargs: Any,
    ) -> Verification:
        analyzer = self._analyze_headers(header)
        verifications = [
            describe(analyzer, **kwargs)
            for describe in self._headers_descriptions
        ]
        status = merge_statuses(v.status for v in verifications)
        return Verification(status=status, children=verifications)



"""Scenario compilation."""

from collections.abc import Mapping
from typing import Any, Optional

from preacher.core.scenario import Scenario
from .description import DescriptionCompiler
from .error import CompilationError, NamedNode
from .case import CaseCompiler
from .response_description import ResponseDescriptionCompiler
from .util import map_on_key, run_on_key


_KEY_LABEL = 'label'
_KEY_WHEN = 'when'
_KEY_DEFAULT = 'default'
_KEY_CASES = 'cases'
_KEY_SUBSCENARIOS = 'subscenarios'


class ScenarioCompiler:

    def __init__(
        self,
        description_compiler: Optional[DescriptionCompiler] = None,
        case_compiler: Optional[CaseCompiler] = None,
    ):
        self._description_compiler = (
            description_compiler or DescriptionCompiler()
        )
        self._case_compiler = case_compiler or CaseCompiler(
            response_compiler=ResponseDescriptionCompiler(
                description_compiler=self._description_compiler,
            ),
        )

    def compile(self, obj: Any) -> Scenario:
        """`obj` should be a mapping."""

        if not isinstance(obj, Mapping):
            raise CompilationError('Must be a mapping')

        label = obj.get(_KEY_LABEL)
        if label is not None and not isinstance(label, str):
            raise CompilationError(
                message='Must be a string',
                path=[NamedNode(_KEY_LABEL)],
            )

        default = obj.get(_KEY_DEFAULT, {})
        if not isinstance(default, Mapping):
            raise CompilationError(
                message='Must be a mapping',
                path=[NamedNode(_KEY_DEFAULT)],
            )
        case_compiler = run_on_key(
            _KEY_DEFAULT,
            self._case_compiler.of_default,
            default,
        )

        condition_objs = obj.get(_KEY_WHEN, [])
        if not isinstance(condition_objs, list):
            condition_objs = [condition_objs]
        conditions = list(map_on_key(
            _KEY_WHEN,
            self._description_compiler.compile,
            condition_objs,
        ))

        case_objs = obj.get(_KEY_CASES, [])
        if not isinstance(case_objs, list):
            raise CompilationError(
                message='Must be a list',
                path=[NamedNode(_KEY_CASES)],
            )
        cases = list(map_on_key(_KEY_CASES, case_compiler.compile, case_objs))

        subscenario_objs = obj.get(_KEY_SUBSCENARIOS, [])
        if not isinstance(subscenario_objs, list):
            raise CompilationError(
                message='Must be a list',
                path=[NamedNode(_KEY_SUBSCENARIOS)],
            )
        subscenario_compiler = ScenarioCompiler(case_compiler=case_compiler)
        subscenarios = list(map_on_key(
            _KEY_SUBSCENARIOS,
            subscenario_compiler.compile,
            subscenario_objs,
        ))

        return Scenario(
            label=label,
            conditions=conditions,
            cases=cases,
            subscenarios=subscenarios,
        )



from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Iterable, List, Optional

from preacher.core.scenario import ScenarioResult
from preacher.core.status import Status
from preacher.compilation.error import CompilationError
from preacher.compilation.yaml import load
from preacher.compilation.scenario import ScenarioCompiler
from .listener import Listener


class Application:
    def __init__(
        self,
        presentations: List[Listener],
        base_url: str,
        retry: int = 0,
        delay: float = 0.1,
        timeout: Optional[float] = None,
    ):
        self._presentations = presentations
        self._base_url = base_url
        self._retry = retry
        self._delay = delay
        self._timeout = timeout

        self._scenario_compiler = ScenarioCompiler()
        self._is_succeeded = True

    @property
    def is_succeeded(self) -> bool:
        return self._is_succeeded

    def run(
        self,
        executor: ThreadPoolExecutor,
        config_paths: Iterable[str],
    ) -> None:
        tasks = [self._submit_each(executor, path) for path in config_paths]
        results = (task() for task in tasks)
        for result in results:
            self._is_succeeded &= result.status.is_succeeded
            for presentation in self._presentations:
                presentation.accept(result)

    def _submit_each(
        self,
        executor: ThreadPoolExecutor,
        config_path: str,
    ) -> Callable[[], ScenarioResult]:
        try:
            scenario_obj = load(config_path)
            scenario = self._scenario_compiler.compile(scenario_obj)
        except CompilationError as error:
            result = ScenarioResult(
                label=f'Compilation Error ({config_path})',
                status=Status.FAILURE,
                message=str(error),
            )
            return lambda: result

        return scenario.submit(
            executor,
            base_url=self._base_url,
            retry=self._retry,
            delay=self._delay,
            timeout=self._timeout,
        ).result



"""Description compilation."""

from collections.abc import Mapping
from typing import Any, Optional

from preacher.core.description import Description
from .error import CompilationError
from .predicate import PredicateCompiler
from .extraction import ExtractionCompiler
from .util import run_on_key, map_on_key


_KEY_DESCRIBE = 'describe'
_KEY_SHOULD = 'should'


class DescriptionCompiler:

    def __init__(
        self,
        extraction_compiler: Optional[ExtractionCompiler] = None,
        predicate_compiler: Optional[PredicateCompiler] = None,
    ):
        self._extraction_compiler = extraction_compiler or ExtractionCompiler()
        self._predicate_compiler = predicate_compiler or PredicateCompiler()

    def compile(self, obj: Any):
        """`obj` should be a mapping."""

        if not isinstance(obj, Mapping):
            raise CompilationError('Must be a mapping')

        extraction_obj = obj.get(_KEY_DESCRIBE)
        extractor = run_on_key(
            _KEY_DESCRIBE,
            self._extraction_compiler.compile,
            extraction_obj
        )

        predicate_objs = obj.get(_KEY_SHOULD, [])
        if not isinstance(predicate_objs, list):
            predicate_objs = [predicate_objs]
        predicates = list(map_on_key(
            _KEY_SHOULD,
            self._predicate_compiler.compile,
            predicate_objs,
        ))

        return Description(extractor=extractor, predicates=predicates)



"""Request compilation."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Mapping as MappingType, Optional

from preacher.core.request import Request
from .error import CompilationError, NamedNode
from .util import or_default


_KEY_PATH = 'path'
_KEY_HEADERS = 'headers'
_KEY_PARAMS = 'params'


@dataclass(frozen=True)
class _Compiled:
    path: Optional[str] = None
    headers: Optional[MappingType[str, str]] = None
    params: Optional[MappingType[str, Any]] = None

    def updated(self, updater: _Compiled) -> _Compiled:
        return _Compiled(
            path=or_default(updater.path, self.path),
            headers=or_default(updater.headers, self.headers),
            params=or_default(updater.params, self.params),
        )

    def to_request(self) -> Request:
        return Request(
            path=or_default(self.path, ''),
            headers=or_default(self.headers, {}),
            params=or_default(self.params, {}),
        )


def _compile(obj) -> _Compiled:
    """`obj` should be a mapping or a string."""

    if isinstance(obj, str):
        return _compile({_KEY_PATH: obj})

    if not isinstance(obj, Mapping):
        raise CompilationError('Must be a mapping or a string')

    path = obj.get(_KEY_PATH)
    if path is not None and not isinstance(path, str):
        raise CompilationError(
            message='Must be a string',
            path=[NamedNode(_KEY_PATH)],
        )

    headers = obj.get(_KEY_HEADERS)
    if headers is not None and not isinstance(headers, Mapping):
        raise CompilationError(
            message='Must be a mapping',
            path=[NamedNode(_KEY_HEADERS)],
        )

    params = obj.get(_KEY_PARAMS)
    if params is not None and not isinstance(params, Mapping):
        raise CompilationError(
            message='Must be a mapping',
            path=[NamedNode(_KEY_PARAMS)],
        )

    return _Compiled(path=path, headers=headers, params=params)


class RequestCompiler:
    def __init__(self, default: _Compiled = None):
        self._default = default or _Compiled()

    def compile(self, obj) -> Request:
        """`obj` should be a mapping or a string."""

        compiled = _compile(obj)
        return self._default.updated(compiled).to_request()

    def of_default(self, obj) -> RequestCompiler:
        """`obj` should be a mapping or a string."""

        compiled = _compile(obj)
        return RequestCompiler(default=compiled)



from __future__ import annotations

import os
from typing import List

import jinja2

from preacher.core.scenario import ScenarioResult
from . import Listener


class ReportingListener(Listener):

    def __init__(self, path):
        self._path = path

    def __enter__(self) -> ReportingListener:
        self._results: List[ScenarioResult] = []
        os.makedirs(self._path, exist_ok=True)
        return self

    def __exit__(self, ex_type, ex_value, trace) -> None:
        env = jinja2.Environment(
            loader=jinja2.PackageLoader('preacher', 'resources/html'),
            autoescape=jinja2.select_autoescape(['html', 'xml'])
        )

        html_path = os.path.join(self._path, 'index.html')
        with open(html_path, 'w') as f:
            env.get_template('index.html').stream(
                scenarios=self._results,
            ).dump(f)

    def accept(self, result: ScenarioResult) -> None:
        self._results.append(result)



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Any, List, Optional

from preacher.core.analysis import Analysis, analyze_json_str
from preacher.core.body_description import BodyDescription
from preacher.core.description import Description
from .analysis import AnalysisCompiler
from .description import DescriptionCompiler
from .error import CompilationError
from .util import map_on_key, or_default, run_on_key


_KEY_ANALYSIS = 'analyze_as'
_KEY_DESCRIPTIONS = 'descriptions'


@dataclass(frozen=True)
class Compiled:
    analyze: Optional[Analysis] = None
    descriptions: Optional[List[Description]] = None

    def convert(self) -> BodyDescription:
        return BodyDescription(
            analyze=or_default(self.analyze, analyze_json_str),
            descriptions=or_default(self.descriptions, []),
        )


class BodyDescriptionCompiler:

    def __init__(
        self,
        default: Optional[Compiled] = None,
        analysis_compiler: Optional[AnalysisCompiler] = None,
        description_compiler: Optional[DescriptionCompiler] = None,
    ):
        self._default = default or Compiled()
        self._analysis_compiler = analysis_compiler or AnalysisCompiler()
        self._description_compiler = (
            description_compiler or DescriptionCompiler()
        )

    def of_default(
        self,
        default: Optional[Compiled],
    ) -> BodyDescriptionCompiler:
        return BodyDescriptionCompiler(
            default=default,
            analysis_compiler=self._analysis_compiler,
            description_compiler=self._description_compiler,
        )

    def compile(self, obj: Any) -> Compiled:
        """
        `obj` should be a mapping or a list.
        An empty list results in an empty description.
        """

        if isinstance(obj, list):
            return self.compile({_KEY_DESCRIPTIONS: obj})
        if not isinstance(obj, Mapping):
            raise CompilationError('Must be a mapping or a list')

        replacements = {}

        analyze_obj = obj.get(_KEY_ANALYSIS)
        if analyze_obj is not None:
            replacements['analyze'] = run_on_key(
                _KEY_ANALYSIS,
                self._analysis_compiler.compile,
                analyze_obj,
            )

        descs_obj = obj.get(_KEY_DESCRIPTIONS)
        if descs_obj is not None:
            replacements['descriptions'] = (  # type: ignore
                self._compile_descriptions(descs_obj)
            )

        return replace(self._default, **replacements)

    def _compile_descriptions(self, obj: Any) -> List[Description]:
        if not isinstance(obj, list):
            obj = [obj]

        return list(map_on_key(
            _KEY_DESCRIPTIONS,
            self._description_compiler.compile,
            obj,
        ))



import time

from typing import Callable, TypeVar


T = TypeVar('T')


def retry_while_false(func: Callable, attempts: int = 1, delay: float = 0.1):
    if attempts < 1:
        raise ValueError(f'`attempts` must be positive, given {attempts}')

    for _ in range(attempts - 1):
        result = func()
        if result:
            return result
        time.sleep(delay)

    return func()



"""Extraction compilation."""

from collections.abc import Mapping

from preacher.core.extraction import (
    Cast,
    Extractor,
    JqExtractor,
    XPathExtractor,
)
from .error import CompilationError, NamedNode
from .util import run_on_key


_EXTRACTION_MAP = {
    'jq': JqExtractor,
    'xpath': XPathExtractor,
}
_CAST_FUNC_MAP = {
    'int': int,
    'float': float,
    'string': str,
}
_EXTRACTION_KEYS = frozenset(_EXTRACTION_MAP.keys())
_KEY_MULTIPLE = 'multiple'
_KEY_CAST_TO = 'cast_to'


class ExtractionCompiler:

    def compile(self, obj) -> Extractor:
        """`obj` should be a mapping or a string."""

        if isinstance(obj, str):
            return self.compile({'jq': obj})

        if not isinstance(obj, Mapping):
            raise CompilationError('Must be a mapping or a string')

        keys = _EXTRACTION_KEYS.intersection(obj.keys())
        if len(keys) != 1:
            raise CompilationError(
                f'Extraction must have only 1 valid key, but has {len(keys)}'
            )
        key = next(iter(keys))

        func = _EXTRACTION_MAP[key]
        query = obj[key]

        multiple = obj.get(_KEY_MULTIPLE, False)
        if not isinstance(multiple, bool):
            raise CompilationError(
                message='Must be a boolean',
                path=[NamedNode(_KEY_MULTIPLE)],
            )

        cast = None
        cast_obj = obj.get(_KEY_CAST_TO)
        if cast_obj is not None:
            cast = run_on_key(_KEY_CAST_TO, self._compile_cast, cast_obj)

        return func(query, multiple=multiple, cast=cast)  # type: ignore

    def _compile_cast(self, obj) -> Cast:
        if not isinstance(obj, str):
            raise CompilationError('Must be a string')

        cast = _CAST_FUNC_MAP.get(obj)  # type: ignore
        if not cast:
            raise CompilationError(f'Invalid value: {obj}')

        return cast



"""Extraction."""

from typing import Any, Callable, List, Optional, Union, TypeVar

import pyjq as jq
from lxml.etree import _Element as Element

from .analysis import Analyzer


T = TypeVar('T')
Cast = Callable[[Optional[Any]], Optional[Any]]


def _default_cast(value: T) -> T:
    return value


class JqExtractor:

    def __init__(
        self,
        query: str,
        multiple: bool = False,
        cast: Optional[Cast] = None,
    ):
        self._query = query
        self._multiple = multiple
        self._cast = cast or _default_cast

        compiled = jq.compile(self._query)
        self._jq = compiled.all

    def extract(self, analyzer: Analyzer) -> Optional[Any]:
        values = (
            self._cast(value) if value is not None else value
            for value in analyzer.jq(self._jq)
        )
        if self._multiple:
            return list(values)
        else:
            return next(values, None)


class XPathExtractor:

    def __init__(
        self,
        query: str,
        multiple: bool = False,
        cast: Optional[Cast] = None,
    ):
        self._query = query
        self._multiple = multiple
        self._cast = cast or _default_cast

    def extract(self, analyzer: Analyzer) -> Optional[Any]:
        elems = analyzer.xpath(self._extract)
        if not elems:
            return None

        values = (self._cast(self._convert(elem)) for elem in elems)
        if self._multiple:
            return list(values)
        else:
            return next(values, None)

    def _convert(self, elem: Element) -> str:
        if isinstance(elem, Element):
            return elem.text
        return str(elem)

    def _extract(self, elem: Element) -> List[Any]:
        return elem.xpath(self._query)


Extractor = Union[JqExtractor, XPathExtractor]



import json
from typing import Any, Callable, TypeVar, Union

from lxml.etree import _Element as Element, XMLParser, fromstring


T = TypeVar('T')


class JsonAnalyzer:

    def __init__(self, json_body: Any):
        self._json_body = json_body

    def jq(self, extract: Callable[[Any], T]) -> T:
        return extract(self._json_body)

    def xpath(self, extract: Callable[[Element], T]) -> T:
        raise NotImplementedError('XPath extraction is not allowed for JSON')


class XmlAnalyzer:

    def __init__(self, etree: Element):
        self._etree = etree

    def jq(self, extract: Callable[[Any], T]) -> T:
        raise NotImplementedError('jq extraction is not allowed for XML')

    def xpath(self, extract: Callable[[Element], T]) -> T:
        return extract(self._etree)


Analyzer = Union[JsonAnalyzer, XmlAnalyzer]
Analysis = Callable[[str], Analyzer]


def analyze_json_str(value: str) -> Analyzer:
    return JsonAnalyzer(json.loads(value))


def analyze_xml_str(value: str) -> Analyzer:
    etree = fromstring(value, parser=XMLParser())
    return XmlAnalyzer(etree)



from typing import Any

from preacher.core.analysis import Analysis, analyze_json_str, analyze_xml_str
from .error import CompilationError


_ANALYSIS_MAP = {
    'json': analyze_json_str,
    'xml': analyze_xml_str,
}


class AnalysisCompiler:

    def compile(self, obj: Any) -> Analysis:
        """`obj` should be a string."""
        if not isinstance(obj, str):
            raise CompilationError('Must be a string')

        analysis = _ANALYSIS_MAP.get(obj)
        if not analysis:
            raise CompilationError(f'Invalid key: {obj}')

        return analysis



from __future__ import annotations

import os
from collections.abc import Mapping
from functools import partial
from typing import Any, Optional, Union

from ruamel.yaml import YAML, Node
from ruamel.yaml.constructor import ConstructorError

from .error import CompilationError
from .util import map, run_on_key


PathLike = Union[str, os.PathLike]


class _Inclusion:
    yaml_tag = '!include'

    def __init__(self, path: Optional[Any]):
        self._path = path

    def resolve(self, origin: PathLike, yaml: YAML) -> Optional[Any]:
        if not isinstance(self._path, str):
            raise CompilationError('Must be a string')

        path = os.path.join(os.path.dirname(origin), self._path)
        return _load(path, yaml)

    @classmethod
    def from_yaml(cls, constructor, node: Node) -> _Inclusion:
        return _Inclusion(node.value)


def _resolve(obj: Any, origin: PathLike, yaml: YAML) -> Optional[Any]:
    resolve = partial(_resolve, origin=origin, yaml=yaml)

    if isinstance(obj, Mapping):
        return {k: run_on_key(k, resolve, v) for (k, v) in obj.items()}

    if isinstance(obj, list):
        return list(map(resolve, obj))

    if isinstance(obj, _Inclusion):
        return obj.resolve(origin, yaml)

    return obj


def _load(path: PathLike, yaml: YAML) -> Optional[Any]:
    with open(path) as f:
        try:
            obj = yaml.load(f)
        except ConstructorError as error:
            raise CompilationError(message=str(error), cause=error)

        return _resolve(obj, path, yaml)


def load(path: PathLike) -> Optional[Any]:
    yaml = YAML(typ='safe', pure=True)
    yaml.register_class(_Inclusion)
    return _load(path, yaml)



from __future__ import annotations

from preacher.core.scenario import ScenarioResult
from . import Listener


class EmptyListener(Listener):

    def __enter__(self) -> EmptyListener:
        return self

    def __exit__(self, ex_type, ex_value, trace) -> None:
        pass

    def accept(self, result: ScenarioResult) -> None:
        pass



"""Response description compilations."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Any, Optional, List

from preacher.core.description import Description, Predicate
from preacher.core.response_description import ResponseDescription
from .body_description import BodyDescriptionCompiler, Compiled as BodyCompiled
from .description import DescriptionCompiler
from .error import CompilationError
from .predicate import PredicateCompiler
from .util import map_on_key, or_default, run_on_key


_KEY_STATUS_CODE = 'status_code'
_KEY_HEADERS = 'headers'
_KEY_BODY = 'body'


@dataclass(frozen=True)
class Compiled:
    status_code: Optional[List[Predicate]] = None
    headers: Optional[List[Description]] = None
    body: Optional[BodyCompiled] = None

    def convert(self) -> ResponseDescription:
        return ResponseDescription(
            status_code_predicates=or_default(self.status_code, []),
            headers_descriptions=or_default(self.headers, []),
            body_description=self.body.convert() if self.body else None,
        )


class ResponseDescriptionCompiler:

    def __init__(
        self,
        default: Optional[Compiled] = None,
        predicate_compiler: Optional[PredicateCompiler] = None,
        description_compiler: Optional[DescriptionCompiler] = None,
        body_description_compiler: Optional[BodyDescriptionCompiler] = None,
    ):
        self._default = default or Compiled()
        self._predicate_compiler = predicate_compiler or PredicateCompiler()
        self._description_compiler = (
            description_compiler
            or DescriptionCompiler(predicate_compiler=self._predicate_compiler)
        )
        self._body_description_compiler = (
            body_description_compiler
            or BodyDescriptionCompiler(
                description_compiler=self._description_compiler
            )
        )

    def of_default(self, default: Compiled) -> ResponseDescriptionCompiler:
        return ResponseDescriptionCompiler(
            default=default,
            predicate_compiler=self._predicate_compiler,
            description_compiler=self._description_compiler,
            body_description_compiler=(
                self._body_description_compiler.of_default(default.body)
            ),
        )

    def compile(self, obj: Any) -> Compiled:
        """`obj` should be a mapping."""

        if not isinstance(obj, Mapping):
            raise CompilationError('Must be a mapping')

        replacements = {}

        status_code_obj = obj.get(_KEY_STATUS_CODE)
        if status_code_obj is not None:
            replacements['status_code'] = (
                self._compile_status_code(status_code_obj)
            )

        headers_obj = obj.get(_KEY_HEADERS)
        if headers_obj is not None:
            replacements['headers'] = run_on_key(  # type: ignore
                _KEY_HEADERS,
                self._compile_headers,
                headers_obj,
            )

        body_obj = obj.get(_KEY_BODY)
        if body_obj is not None:
            replacements['body'] = run_on_key(  # type: ignore
                _KEY_BODY,
                self._body_description_compiler.compile,
                body_obj,
            )

        return replace(self._default, **replacements)

    def _compile_status_code(self, obj: Any) -> List[Predicate]:
        if not isinstance(obj, list):
            obj = [obj]
        return list(map_on_key(
            _KEY_STATUS_CODE,
            self._predicate_compiler.compile,
            obj,
        ))

    def _compile_headers(self, obj: Any) -> List[Description]:
        if isinstance(obj, Mapping):
            obj = [obj]
        if not isinstance(obj, list):
            message = 'Must be a list or a mapping'
            raise CompilationError(message=message)
        return list(map_on_key(
            _KEY_HEADERS,
            self._description_compiler.compile,
            obj,
        ))



"""Case compilation."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Optional

from preacher.core.case import Case
from .error import CompilationError, NamedNode
from .request import RequestCompiler
from .response_description import ResponseDescriptionCompiler
from .util import run_on_key


_KEY_LABEL = 'label'
_KEY_REQUEST = 'request'
_KEY_RESPONSE = 'response'


class CaseCompiler:

    def __init__(
        self,
        request_compiler: Optional[RequestCompiler] = None,
        response_compiler: Optional[ResponseDescriptionCompiler] = None
    ):
        self._request_compiler = request_compiler or RequestCompiler()
        self._response_compiler = (
            response_compiler or ResponseDescriptionCompiler()
        )

    @property
    def request_compiler(self) -> RequestCompiler:
        return self._request_compiler

    def compile(self, obj: Any) -> Case:
        """`obj` should be a mapping."""

        if not isinstance(obj, Mapping):
            raise CompilationError('Must be a mapping')

        label = obj.get(_KEY_LABEL)
        if label is not None and not isinstance(label, str):
            raise CompilationError(
                message=f'Case.{_KEY_LABEL} must be a string',
                path=[NamedNode(_KEY_LABEL)],
            )

        request = run_on_key(
            _KEY_REQUEST,
            self._request_compiler.compile,
            obj.get(_KEY_REQUEST, {}),
        )
        response_description = run_on_key(
            _KEY_RESPONSE,
            self._response_compiler.compile,
            obj.get(_KEY_RESPONSE, {}),
        ).convert()
        return Case(
            label=label,
            request=request,
            response_description=response_description,
        )

    def of_default(self, obj: Mapping) -> CaseCompiler:
        request_compiler = run_on_key(
            _KEY_REQUEST,
            self._request_compiler.of_default,
            obj.get(_KEY_REQUEST, {}),
        )

        res_compiled = run_on_key(
            _KEY_RESPONSE,
            self._response_compiler.compile,
            obj.get(_KEY_RESPONSE, {}),
        )
        res_compiler = self._response_compiler.of_default(res_compiled)
        return CaseCompiler(
            request_compiler=request_compiler,
            response_compiler=res_compiler,
        )



"""Description."""

from typing import Any, Callable, List

from .analysis import Analyzer
from .extraction import Extractor
from .status import merge_statuses
from .verification import Verification


Predicate = Callable


class Description:

    def __init__(self, extractor: Extractor, predicates: List[Predicate]):
        self._extractor = extractor
        self._predicates = predicates

    def __call__(self, analyzer: Analyzer, **kwargs: Any) -> Verification:
        """`**kwargs` will be delegated to predicates."""
        try:
            verified_value = self._extractor.extract(analyzer)
        except Exception as error:
            return Verification.of_error(error)

        verifications = [
            predicate(verified_value, **kwargs)
            for predicate in self._predicates
        ]
        status = merge_statuses(v.status for v in verifications)
        return Verification(status, children=verifications)

    @property
    def extractor(self) -> Extractor:
        return self._extractor

    @property
    def predicates(self) -> List[Predicate]:
        return self._predicates



"""Request."""

from copy import copy
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Mapping, Optional

import requests

from .datetime import now

from preacher import __version__ as _VERSION


_DEFAULT_HEADERS = {'User-Agent': f'Preacher {_VERSION}'}


@dataclass(frozen=True)
class Response:
    status_code: int
    headers: Mapping[str, str]
    body: str
    request_datetime: datetime


class Request:
    def __init__(
        self,
        path: str = '',
        headers: Mapping[str, str] = {},
        params: Mapping[str, Any] = {},
    ):
        self._path = path
        self._headers = headers
        self._params = params

    def __call__(
        self,
        base_url: str,
        timeout: Optional[float] = None,
    ) -> Response:
        headers = copy(_DEFAULT_HEADERS)
        headers.update(self._headers)
        request_datetime = now()

        res = requests.get(
            base_url + self._path,
            headers=headers,
            params=self._params,
            timeout=timeout,
        )

        return Response(
            status_code=res.status_code,
            headers={
                # Convert to the normal dictionary to adapt jq.
                # Names are converted to lower case to normalize.
                name.lower(): value for (name, value) in res.headers.items()
            },
            body=res.text,
            request_datetime=request_datetime
        )

    @property
    def path(self) -> str:
        return self._path

    @property
    def headers(self) -> Mapping:
        return self._headers

    @property
    def params(self) -> Mapping:
        return self._params



"""Status."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from functools import reduce, singledispatch
from typing import (
    Iterable as IterableType,
    Iterator,
    Sequence,
    TypeVar,
    Union,
)


T = TypeVar('T')


class Status(Enum):
    """
    >>> Status.SKIPPED.is_succeeded
    True
    >>> bool(Status.SKIPPED)
    True
    >>> Status.SKIPPED.merge(Status.SUCCESS)
    SUCCESS
    >>> Status.SKIPPED.merge(Status.UNSTABLE)
    UNSTABLE
    >>> Status.SKIPPED.merge(Status.FAILURE)
    FAILURE

    >>> Status.SUCCESS.is_succeeded
    True
    >>> bool(Status.SUCCESS)
    True
    >>> Status.SUCCESS.merge(Status.SKIPPED)
    SUCCESS
    >>> Status.SUCCESS.merge(Status.SUCCESS)
    SUCCESS
    >>> Status.SUCCESS.merge(Status.UNSTABLE)
    UNSTABLE
    >>> Status.SUCCESS.merge(Status.FAILURE)
    FAILURE

    >>> Status.UNSTABLE.is_succeeded
    False
    >>> bool(Status.UNSTABLE)
    False
    >>> Status.UNSTABLE.merge(Status.SKIPPED)
    UNSTABLE
    >>> Status.UNSTABLE.merge(Status.SUCCESS)
    UNSTABLE
    >>> Status.UNSTABLE.merge(Status.UNSTABLE)
    UNSTABLE
    >>> Status.UNSTABLE.merge(Status.FAILURE)
    FAILURE

    >>> Status.FAILURE.is_succeeded
    False
    >>> bool(Status.FAILURE)
    False
    >>> Status.FAILURE.merge(Status.SKIPPED)
    FAILURE
    >>> Status.FAILURE.merge(Status.SUCCESS)
    FAILURE
    >>> Status.FAILURE.merge(Status.UNSTABLE)
    FAILURE
    >>> Status.FAILURE.merge(Status.FAILURE)
    FAILURE
    """
    # Numbers stand for the priorities for merging.
    SKIPPED = 0
    SUCCESS = 1
    UNSTABLE = 2
    FAILURE = 3

    @property
    def is_succeeded(self):
        return self.value <= Status.SUCCESS.value

    def merge(self, other: Status):
        return max(self, other, key=lambda status: status.value)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return str(self)

    def __bool__(self) -> bool:
        return self.is_succeeded


@singledispatch
def merge_statuses(*args) -> Status:
    """
    >>> merge_statuses(1)
    Traceback (most recent call last):
        ...
    ValueError: (1,)

    For varargs.
    >>> merge_statuses(Status.UNSTABLE)
    UNSTABLE
    >>> merge_statuses(Status.SUCCESS, Status.FAILURE, Status.UNSTABLE)
    FAILURE

    For iterables.
    >>> merge_statuses([])
    SKIPPED
    >>> merge_statuses([Status.SUCCESS, Status.UNSTABLE, Status.FAILURE])
    FAILURE
    """
    raise ValueError(str(args))


@merge_statuses.register
def _merge_statuses_for_varargs(*statuses: Status):
    return merge_statuses(statuses)


@merge_statuses.register
def _merge_statuses_for_iterable(statuses: Iterable):
    return reduce(lambda lhs, rhs: lhs.merge(rhs), statuses, Status.SKIPPED)


@dataclass(frozen=True)
class StatusedMixin:
    status: Status = Status.SKIPPED


class StatusedInterface(ABC):
    """
    >>> class ConcreteStatused(StatusedInterface):
    ...     @property
    ...     def status(self) -> Status:
    ...         return super().status
    >>> ConcreteStatused().status
    SKIPPED
    """
    @property
    @abstractmethod
    def status(self) -> Status:
        return Status.SKIPPED


class StatusedSequence(StatusedInterface, Sequence[T]):
    """
    >>> bool(StatusedSequence())
    False
    >>> bool(StatusedSequence(Status.SUCCESS, []))
    False
    >>> bool(StatusedSequence(Status.FAILURE, [1]))
    True
    """

    def __init__(
        self,
        status: Status = Status.SKIPPED,
        items: Sequence[T] = [],
    ):
        self._status = status
        self._items = items

    @property
    def status(self) -> Status:
        return self._status

    def __bool__(self) -> bool:
        return bool(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> Iterator[T]:
        return iter(self._items)

    def __getitem__(self, key):
        return self._items[key]


Statused = Union[StatusedMixin, StatusedInterface]
StatusedType = TypeVar('StatusedType', bound=Statused)


def collect_statused(
    items: IterableType[StatusedType],
) -> StatusedSequence[StatusedType]:
    items = list(items)
    status = merge_statuses(item.status for item in items)
    return StatusedSequence(status=status, items=items)



from __future__ import annotations

from abc import ABC, abstractmethod

from preacher.core.scenario import ScenarioResult


class Listener(ABC):

    @abstractmethod
    def __enter__(self) -> Listener:
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, ex_type, ex_value, trace) -> None:
        raise NotImplementedError()

    @abstractmethod
    def accept(self, result: ScenarioResult) -> None:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['preacher',
 'preacher.app',
 'preacher.app.cli',
 'preacher.app.listener',
 'preacher.compilation',
 'preacher.core']

package_data = \
{'': ['*'], 'preacher': ['resources/html/*', 'resources/html/macros/*']}

install_requires = \
['aniso8601>=8.0,<9.0',
 'jinja2>=2.10,<3.0',
 'lxml>=4.4,<5.0',
 'pyhamcrest>=1.9,<2.0',
 'pyjq>=2.3,<3.0',
 'requests>=2.21,<3.0',
 'ruamel.yaml>=0.15.89,<0.16.0']

entry_points = \
{'console_scripts': ['preacher-cli = preacher.app.cli.main:main']}

setup_kwargs = {
    'name': 'preacher',
    'version': '0.11.0',
    'description': 'A web API verification tool.',
    'long_description': '# Preacher: Flexible Web API Verification\n\n[![PyPI version](https://badge.fury.io/py/preacher.svg)](https://badge.fury.io/py/preacher)\n[![Documentation Status](https://readthedocs.org/projects/preacher/badge/?version=latest)](https://preacher.readthedocs.io/en/latest/?badge=latest)\n[![CircleCI](https://circleci.com/gh/ymoch/preacher.svg?style=svg)](https://circleci.com/gh/ymoch/preacher)\n[![codecov](https://codecov.io/gh/ymoch/preacher/branch/master/graph/badge.svg)](https://codecov.io/gh/ymoch/preacher)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/ymoch/preacher.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/ymoch/preacher/context:python)\n\nPreacher verifies API servers,\nwhich requests to the servers and verify the responses along to given scenarios.\n\nScenarios are written in [YAML][], bodies are analyzed [jq][] or [XPath][] queries\nand validation rules are based on [Hamcrest][] ([PyHamcrest][])\nso that any developers can write without learning toughly.\n\nThe full documentation is available at\n[preacher.readthedocs.io](https://preacher.readthedocs.io/).\n\n## Targets\n\n- Flexible validation to test with real backends: neither mocks nor sandboxes.\n  - Matcher-based validation.\n- CI Friendly to automate easily.\n  - A CLI application and YAML-based scenarios.\n\n## Usage\n\nFirst, install from PyPI. Supports only Python 3.7+.\n\n```sh\n$ pip install preacher\n```\n\nSecond, write your own scenario.\n\n```yaml\n# scenario.yml\nlabel: An example of a scenario\ncases:\n  - label: An example of a case\n    request: /path/to/foo\n    response:\n      status_code: 200\n      body:\n        - describe: .foo\n          should:\n            equal: bar\n```\n\nThen, run ``preacher-cli`` command.\n\n```sh\n$ preacher-cli -u http://your.domain.com/base scenario.yml\n```\n\nFor more information such as grammer of scenarios,\nsee [the full documentation](https://preacher.readthedocs.io/).\n\n## License\n\n[![MIT License](https://img.shields.io/badge/License-MIT-brightgreen.svg)](https://opensource.org/licenses/MIT)\n\nCopyright (c) 2019 Yu MOCHIZUKI\n\n[YAML]: https://yaml.org/\n[jq]: https://stedolan.github.io/jq/\n[XPath]: https://www.w3.org/TR/xpath/all/\n[Hamcrest]: http://hamcrest.org/\n[PyHamcrest]: https://pyhamcrest.readthedocs.io/\n',
    'author': 'Yu Mochizuki',
    'author_email': 'ymoch.dev@gmail.com',
    'url': 'https://preacher.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dataclasses import asdict, dataclass

from .analysis import Analyzer, JsonAnalyzer


@dataclass
class Context:
    base_url: str = ''

    def analyze(self) -> Analyzer:
        return JsonAnalyzer(asdict(self))



"""Preacher CLI."""

from __future__ import annotations

import argparse
import logging
import sys
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

from preacher import __version__ as VERSION
from preacher.app.application import Application
from preacher.app.listener import Listener
from preacher.app.listener.empty import EmptyListener
from preacher.app.listener.logging import LoggingListener
from preacher.app.listener.report import ReportingListener


DEFAULT_URL = 'http://localhost:5042'
DEFAULT_URL_DESCRIPTION = 'the sample'

HANDLER = logging.StreamHandler()
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(HANDLER)


LOGGING_LEVEL_MAP = {
    'skipped': logging.DEBUG,
    'success': logging.INFO,
    'unstable': logging.WARN,
    'failure': logging.ERROR,
}


def report_to(path: Optional[str] = None) -> Listener:
    if not path:
        return EmptyListener()
    return ReportingListener(path)


def zero_or_positive_int(value: str) -> int:
    int_value = int(value)
    if int_value < 0:
        raise argparse.ArgumentTypeError(
            f"must be positive or 0, given {int_value}"
        )
    return int_value


def positive_float(value: str) -> float:
    float_value = float(value)
    if float_value <= 0.0:
        raise argparse.ArgumentTypeError(
            f"must be positive, given {float_value}"
        )
    return float_value


def zero_or_positive_float(value: str) -> float:
    float_value = float(value)
    if float_value < 0.0:
        raise argparse.ArgumentTypeError(
            f"must be positive or 0, given {float_value}"
        )
    return float_value


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'scenario',
        nargs='+',
        help='scenario file paths'
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=VERSION,
    )
    parser.add_argument(
        '-u', '--url',
        metavar='url',
        help='specify the base URL',
        default='http://localhost:5042',
    )
    parser.add_argument(
        '-l', '--level',
        choices=LOGGING_LEVEL_MAP.keys(),
        help='show only above or equal to this level',
        default='success',
    )
    parser.add_argument(
        '-r', '--retry',
        type=zero_or_positive_int,
        metavar='num',
        help='set the max retry count',
        default=0,
    )
    parser.add_argument(
        '-d', '--delay',
        type=zero_or_positive_float,
        metavar='sec',
        help='set the delay between attempts in seconds',
        default=0.1,
    )
    parser.add_argument(
        '-t', '--timeout',
        type=positive_float,
        metavar='sec',
        help='set the request timeout in seconds',
    )
    parser.add_argument(
        '-c', '--concurrency',
        type=int,
        metavar='num',
        help='set the request concurrency',
        default=1,
    )
    parser.add_argument(
        '-R', '--report',
        metavar='dir',
        help='report directory (experimental)',
    )

    return parser.parse_args()


def main() -> None:
    """Main."""
    args = parse_args()

    level = LOGGING_LEVEL_MAP[args.level]
    HANDLER.setLevel(level)
    LOGGER.setLevel(level)

    with ThreadPoolExecutor(args.concurrency) as executor, \
            LoggingListener(LOGGER) as logging_listener, \
            report_to(args.report) as reporting_listener:
        app = Application(
            presentations=[logging_listener, reporting_listener],
            base_url=args.url,
            retry=args.retry,
            delay=args.delay,
            timeout=args.timeout,
        )
        app.run(executor, args.scenario)

    if not app.is_succeeded:
        sys.exit(1)



from __future__ import annotations

from typing import Any, List

from .analysis import Analysis, analyze_json_str
from .description import Description
from .status import merge_statuses
from .verification import Verification


class BodyDescription:

    def __init__(
        self,
        analyze: Analysis = analyze_json_str,
        descriptions: List[Description] = [],
    ):
        self._analyze = analyze
        self._descriptions = descriptions

    def verify(self, body: str, **kwargs: Any) -> Verification:
        try:
            analyzer = self._analyze(body)
        except Exception as error:
            return Verification.of_error(error)

        verifications = [
            describe(analyzer, **kwargs)
            for describe in self._descriptions
        ]
        status = merge_statuses(v.status for v in verifications)
        return Verification(status=status, children=verifications)

    @property
    def descriptions(self) -> List[Description]:
        return self._descriptions
